/*
 * Copyright (C) 2005 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License.
 * You must obey the GNU General Public License in all respects for all of the code used
 * other than the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package  mir.entity;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import mir.config.MirPropertiesConfiguration;
import mir.log.LoggerWrapper;
import mir.storage.DatabaseExc;
import mir.storage.DatabaseFailure;
import mir.storage.Database;

/**
 * Base class the entities are derived from. Provides base functionality of
 * an entity.
 *
 * @version $Id: AbstractEntity.java,v 1.8.2.9 2005/12/24 11:43:34 zapata Exp $
 */

public class AbstractEntity implements Entity {
  protected static MirPropertiesConfiguration configuration = MirPropertiesConfiguration.instance();

  protected Map values;
  protected Database database;

  public AbstractEntity() {
    values = new HashMap();
  }

  public void setStorage(Database aDatabase) {
    database = aDatabase;
  }

  /** {@inheritDoc} */
  public void setFieldValues(Map aMap) {
    if (aMap!=null) {
      Iterator i = aMap.entrySet().iterator();
      synchronized(this) {
        while (i.hasNext()) {
          Map.Entry entry = (Map.Entry) i.next();

          setFieldValue( (String) entry.getKey(), (String) entry.getValue());
        }
      }
    }
  }

  /** {@inheritDoc} */
  public String getId() {
    return getFieldValue(database.getIdFieldName());
  }

  /** {@inheritDoc} */
  public void setId(String id) {
    setFieldValue(database.getIdFieldName(), id);
  }

  /** {@inheritDoc} */
  public String insert() throws DatabaseExc {
    getLogger().debug("Entity: trying to insert ...");

    if (database != null) {
      return database.insert(this);
    }

    throw new DatabaseExc("database == null!");
  }

  /** {@inheritDoc} */
  public void update() throws DatabaseFailure {
    database.update(this);
  }

  /** {@inheritDoc} */
  public String getFieldValue(String aFieldName) {
    String returnValue = null;

    if (aFieldName != null) {
      returnValue = (String) values.get(aFieldName);
    }
    return returnValue;
  }

  /** {@inheritDoc} */
  public boolean hasFieldValue(String aFieldName) {
    return values.containsKey(aFieldName);
  }

  /**
   * Sets the value for a field. Issues a log message if the field name
   * supplied was not found in the Entity.
   * @param theProp The field name whose value has to be set
   * @param theValue The new value of the field
   * @exception DatabaseFailure
   */
  public void setFieldValue(String theProp, String theValue) throws DatabaseFailure {
    if (hasField(theProp))
      values.put(theProp, theValue);
    else {
      getLogger().warn("Entity.setFieldValue: Property not found: " + theProp + " (" + theValue + ")");
    }
  }

  /**
   * Returns the field names of the Entity
   */
  public List getFieldNames() throws DatabaseFailure {
    return database.getFieldNames();
  }

  /** Returns whether fieldName is a valid field name of this Entity.
   * @param fieldName
   * @return true in case fieldName is a field name, else false.
   * @exception DatabaseFailure
   */
  public boolean hasField(String fieldName) throws DatabaseFailure {
    return getFieldNames().contains(fieldName);
  }

  protected LoggerWrapper getLogger() {
    return new LoggerWrapper("Entity");
  }
}

